//
//	CColorPicker.cp
//
//	A wrapper for the Color Picker.
//	Requires that the ColorPickerLib be added to the project.
//	It's best to "Import 'Weak'" the ColorPickerLib.
//
//	by James Jennings
//	Started March 7, 1997
//
// CMApplication.h

#include "CColorPicker.h"
#include <ColorPicker.h>
#include <String_Utils.h>

Boolean	CColorPicker::IsAvailable()
{
	OSErr err = noErr;
	Int32 version;
	
	err = ::Gestalt(gestaltColorPickerVersion,&version);
	
	return err == noErr && version >= 0x00000100;
}

Boolean	CColorPicker::HasVersion2()
{
	OSErr err = noErr;
	Int32 version;
	
	err = ::Gestalt(gestaltColorPickerVersion,&version);
	
	return err == noErr && version >= 0x00000200;
}

CColorPicker::CColorPicker()
{
	// Initialize the ColorPickerInfo
	/* setting input color to be an RGB color */
	const CMRGBColor kWhite = { 0xFFFF, 0xFFFF, 0xFFFF };
	mInfo.theColor.color.rgb = kWhite;
	mInfo.theColor.profile = 0L;
	/* no destination profile */
	mInfo.dstProfile = 0L;
	mInfo.flags = CanModifyPalette | CanAnimatePalette;
	/* center dialog box on the deepest color screen */
	mInfo.placeWhere = kDeepestColorScreen;
	/* use the system default picker */
	mInfo.pickerType = 0L;
	/* install event filter and color-changed functions */
	mInfo.eventProc = NewUserEventProc(PickerEventFilterFunction);	/* see Listing 2-2(AC) */
	mInfo.colorProc = nil;							/* see Listing 2-3(AC) */
	mInfo.colorProcData = 0L;
	mInfo.prompt[0] = '\0';
	/* describe the Edit menu for Color Picker Manager */
	// Using PowerPlant hooks, of course.
	MenuHandle mh;
	Int16 menuID;
	LMenuBar::GetCurrentMenuBar()->FindMenuItem(cmd_Cut,   menuID, mh, mInfo.mInfo.cutItem);
	LMenuBar::GetCurrentMenuBar()->FindMenuItem(cmd_Copy,  menuID, mh, mInfo.mInfo.copyItem);
	LMenuBar::GetCurrentMenuBar()->FindMenuItem(cmd_Paste, menuID, mh, mInfo.mInfo.pasteItem);
	LMenuBar::GetCurrentMenuBar()->FindMenuItem(cmd_Clear, menuID, mh, mInfo.mInfo.clearItem);
	LMenuBar::GetCurrentMenuBar()->FindMenuItem(cmd_Undo,  menuID, mh, mInfo.mInfo.undoItem);
	mInfo.mInfo.editMenuID = menuID;
}

CColorPicker::~CColorPicker()
{
	if (mInfo.eventProc)
		DisposeRoutineDescriptor(mInfo.eventProc);
}

void	CColorPicker::SetPrompt( Str255 inPrompt )
{
	if ( inPrompt == nil ) {
		mInfo.prompt[0] = '\0';
	} else {
		CopyPStr(inPrompt, mInfo.prompt);
	}
}

Boolean	CColorPicker::Choose(RGBColor &ioColor)
{
	/* display dialog box to allow user to choose a color */
	
	// You will want to call CColorPicker::IsAvailable() yourself before
	// you allow the user to request a color picker.
	Assert_( IsAvailable() );
	
	// Since the picker allows us to use the Undo item,
	// we need to make it say "Undo". 
	// We don't need to save and restore the old text as PowerPlant
	// will fix that on the next FindCommandStatus() cycle..
	MenuHandle mh = ::GetMenuHandle(mInfo.mInfo.editMenuID);
	Assert_(mh != nil);
	LStr255 str(STRx_UndoEdit, str_RedoUndo);	// undo edit string
	::SetMenuItemText(mh, mInfo.mInfo.undoItem, str);
	
	if (HasVersion2())
		return Choose2(ioColor);
	else
		return Choose1(ioColor);
}

Boolean	CColorPicker::Choose1(RGBColor &ioColor)
{
	// Use the version 1 color picker
	
	Point where = { -1, -1 };	// center on best screen
	
	UDesktop::Deactivate();
	Boolean result = ::GetColor(where,mInfo.prompt,&ioColor,&ioColor);
	UDesktop::Activate();
	
	return result;
}

Boolean	CColorPicker::Choose2(RGBColor &ioColor)
{
	
	mInfo.theColor.color.rgb.red   = ioColor.red;
	mInfo.theColor.color.rgb.green = ioColor.green;
	mInfo.theColor.color.rgb.blue  = ioColor.blue;
	
	UDesktop::Deactivate();
	OSErr theErr = ::PickColor(&mInfo);
	UDesktop::Activate();
	Boolean result = (theErr == noErr && mInfo.newColorChosen);
	
	if (result) {
		ioColor.red   = mInfo.theColor.color.rgb.red;
		ioColor.green = mInfo.theColor.color.rgb.green;
		ioColor.blue  = mInfo.theColor.color.rgb.blue;
	}
		
	return result;
}

pascal Boolean CColorPicker::PickerEventFilterFunction (EventRecord *inEvent) 
{  /* returning false sends events to the Color Picker Manager */
	Boolean handled = false;
	switch(inEvent->what) {
	case updateEvt:
	
		// Handle update for other windows
		// Do exactly what LEventDispatcher::EventUpdate() does.
		LWindow	*theWindow = LWindow::FetchWindowObject(
										(WindowPtr) inEvent->message);
		if (theWindow != nil) {
			theWindow->UpdatePort();
			handled = true;
		}
			
	}
	return handled;
}